<?php
#-------------------------------------------------------------------------------
# Copyright (c) 2009 Gomilsek-informatika.
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the GNU Public License v2.0
# which accompanies this distribution, and is available at
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
# 
# Contributors:
# 	Gomilsek-informatika  (initial API and implementation)
# Contact:
# 	customers@toolsjx.com
#-------------------------------------------------------------------------------

// Check to ensure this file is within the rest of the framework
defined('JPATH_BASE') or die();

/**
 * MySQL database driver
 *
 * @package		Joomla.Framework
 * @subpackage	Database
 * @since		1.0
 */
class JDatabaseMSSQL extends JDatabase
{
	/** @var string The database driver name */
	var $name			= 'mssql';
	/** @var string The database name */
	var $_db			= '';
	/** @var string The null/zero date string */
	var $_nullDate		= '1900-01-01 00:00:00';
	/** @var string Quote for named objects */
	var $_nameQuote		= '[]';
	/** @var array Store Table and Field Information */
	var $_structure		= array( 'fields'=>array(), 'indexes'=>array(), 'keys'=>array(), 'tables'=>array() );
	/** @var array Store ntext fields */
	var $_ntext			= array();
	/** @var array Store date fields */
	var $_dates			= array();
	/** @var array Store numbers of rows */
	var $_numrows		= array();


	function __construct( $options )
	{
		$host		= array_key_exists('host', $options)	? $options['host']		: 'localhost';
		$user		= array_key_exists('user', $options)	? $options['user']		: '';
		$password	= array_key_exists('password',$options)	? $options['password']	: '';
		$database	= array_key_exists('database',$options)	? $options['database']	: '';
		$prefix		= array_key_exists('prefix', $options)	? $options['prefix']	: 'jos_';
		$select		= array_key_exists('select', $options)	? $options['select']	: true;

		// perform a number of fatality checks, then return gracefully
		if (!function_exists( 'mssql_connect' )) {
			$this->_errorNum = 1;
			$this->_errorMsg = 'The MSSQL adapter "mssql" is not available.';
			return;
		} 

		// connect to the server
		if (!($this->_resource = @mssql_connect( $host, $user, $password, true ))) {
			$this->_errorNum = 2;
			$this->_errorMsg = 'Could not connect to MSSQL server ' . $host;
			return;
		}

		// select the database
		if ( $select ) {
			if ( $database != '' && !mssql_select_db( $database, $this->_resource )) {
				$this->_errorNum = 3;
				$this->_errorMsg = 'Could not connect to database';
				return true;
			}
		}

		// finalize initialization
		parent::__construct($options);
	}
	/**
	 * Database object destructor
	 *
	 * @return boolean
	 * @since 1.5
	 */
	function __destruct()
	{
		$return = false;
		if (is_resource($this->_resource)) {
			$return = mssql_close($this->_resource);
		}
		return $return;
	}
	/**
	 * Count the rows affected by the last query (update)
	 * @return int The number of affected rows in the previous operation
	 */
	function getAffectedRows()
	{
		return mssql_rows_affected($this->_resource) - $this->_offset;
	}
	/**
	 * Assumes database collation in use by sampling one text field in one table
	 * @return string Collation in use
	 */
	function getCollation ()
	{
		$cur = mssql_query( 'SELECT CONVERT(sysname,DatabasePropertyEx(db_name(),\'Collation\'))', $this->_resource );
		$arr = mssql_fetch_array($cur);
		return $arr[0];
	}
	/**
	* Get a database escaped string
	* @return string
	*/
	function getEscaped( $text )
	{
		return str_replace( "''", "'", $text );
	}
	/**
	* Count the rows affected by the last query (select)
	* @return int The number of rows returned from the most recent query.
	*/
	function getNumRows( $cur=null )
	{
		$cur = ($cur) ? $cur : $this->_cursor;
		return (isset($this->_numrows[$cur])) ? $this->_numrows[$cur] : null;
	}
	/**
	* Get field names and types for one or more tables
	* @param array 		A list of table names
	* @param boolean 	Whether to reload the fields
	* @return array 	An array of fields by table
	*/
	function getTableFields( $tables, $reload=false )
	{
		$return = array();
		foreach ($this->getTableFieldsExtended( $tables, $reload ) as $tbl=>$fields) {
			foreach ( $fields as $k=>$v ) {
				$return[$tbl][$k] = $v['TYPE_NAME'];
			}
		}
		return $return;
	}
	/**
	* Get table names for the current database
	* @param boolean 	Whether to reload the fields
	* @return array 	A list of table names
	*/
	function getTableList( $reload=false )
	{		
		$return = array();
		foreach ( $this->getTableListExtended( $reload ) as $k=>$v ) {
			$return[] = $k;
		}
		return $return;
	}
	/**
	* Get the SQL server version
	* @return string
	*/
	function getVersion()
	{
		$cur = mssql_query( 'SELECT @@VERSION', $this->_resource );
		$arr = mssql_fetch_array($cur);
		return $arr[0];
	}
	/**
	* Get the ID of the last INSERT row
	* @return mixed	The ID
	*/
	function insertid()
	{
		$cur = mssql_query( 'SELECT SCOPE_IDENTITY()', $this->_resource );
		$arr = mssql_fetch_array($cur);
		return $arr[0];
	}
	/**
	 * Inserts a row into a table based on an objects properties
	 * @param	string	The name of the table
	 * @param	mixed	An object or array whose properties match table fields
	 * @param	string	The name of the primary key. If provided the object property is updated.
	 */
	function insertObject( $table, &$object, $keyName = NULL, $updateDup = false )
	{
		$fields = array();
		$fmtsql = "INSERT INTO $table ( %s ) VALUES ( %s ) ";
		$id = null;
		$identity_insert = 0;
		$array = (array)$object;
		foreach ( $array as $k=>$v ) {
			// 1: NULL or empty Primary Key we will just omit this field and let MSSQL set the ID
			// 2: Value means we have to set IDENTITY_INSERT
			if( $k == $keyName ) {
				if (empty($v)) {
					continue;
				} else {
					$id = $v;
					$identity_insert = 1;
				}
			}
			// 1: Omit Arrays, Objects or Reserved Fields
			// 2: Omit NULL values
			// 3: Include Standard values
			if( is_array($v) or is_object($v) or $k[0] == '_' ) {
				continue;
			} else if ( is_null($v) )	{
				$val = $this->isQuoted( $k ) ? $this->Quote( $v ) : $v;
			} else {
				continue;
			}
			$fields[] = $this->nameQuote( $k );;
			$values[] = $val;
		}
		$this->setQuery( sprintf( $fmtsql, implode( ",", $fields ) ,  implode( ",", $values ) ) );

		// 1: run the query with IDENTITY_INSERT ON, try to update on failure
		// 2: run the query then get the new ID and try to set it
		// 3: faile to run the query
		if ($identity_insert) {
			mssql_query('SET IDENTITY_INSERT ' . $this->nameQuote($table) . ' ON', $this->_resource );
			$cur = $this->query();
			mssql_query('SET IDENTITY_INSERT ' . $this->nameQuote($table) . ' OFF', $this->_resource );
			if ($cur) {
				return true;
			} else if ($updateDup) {
				return $this->updateObject( $table, $object, $keyName );
			} else {
				return false;
			}
		} else if ( $this->query() ) {
			if ( $keyName && $id = $this->insertid() ) {
				if (is_object($object)) {
					$object->$keyName = $id;
				} else if (is_array($object)) {
					$object[$keyName] = $id;
				}
			}
			return true;
		} else {
			return false;
		}
	}
	/**
	* Fetch a result row as an associative array
	*
	* return array
	*/
	function loadAssoc()
	{
		if (!($cur = $this->query())) {
			return null;
		}
		$ret = null;
		if ($array = mssql_fetch_assoc( $cur )) {
			$ret = $array;
		}
		$this->_freeresult( $cur );
		return $ret;
	}
	/**
	* Load a assoc list of database rows
	* @param string The field name of a primary key
	* @return array If <var>key</var> is empty as sequential list of returned records.
	*/
	function loadAssocList( $key='' )
	{
		if (!($cur = $this->query())) {
			return null;
		}
		$array = array();
		while ($row = mssql_fetch_assoc( $cur )) {
			if ($key) {
				$array[$row[$key]] = $row;
			} else {
				$array[] = $row;
			}
		}
		$this->_freeresult( $cur );
		return $array;
	}
	/**
	* This global function loads the first row of a query into an object
	*
	* return object
	*/
	function loadObject( )
	{
		if (!($cur = $this->query())) {
			return null;
		}
		$ret = null;
		if ($object = mssql_fetch_object( $cur )) {
			$ret = $object;
		}
		$this->_freeresult( $cur );
		return $ret;
	}
	/**
	* Load a list of database objects
	* @param string The field name of a primary key
	* @return array If <var>key</var> is empty as sequential list of returned records.
	* If <var>key</var> is not empty then the returned array is indexed by the value
	* the database key.  Returns <var>null</var> if the query fails.
	*/
	function loadObjectList( $key='' )
	{
		if (!($cur = $this->query())) {
			return null;
		}
		$array = array();
		while ($row = mssql_fetch_object( $cur )) {
			if ($key) {
				$array[$row->$key] = $row;
			} else {
				$array[] = $row;
			}
		}
		$this->_freeresult( $cur );
		return $array;
	}
	/**
	* This method loads the first field of the first row returned by the query.
	*
	* @return The value returned in the query or null if the query failed.
	*/
	function loadResult()
	{
		if (!($cur = $this->query())) {
			return null;
		}
		$ret = null;
		if ($row = mssql_fetch_row( $cur )) {
			$ret = $row[0];
		}
		$this->_freeresult( $cur );
		return $ret;
	}
	/**
	* Load an array of single field results into an array
	*/
	function loadResultArray($numinarray = 0)
	{
		if (!($cur = $this->query())) {
			return null;
		}
		$array = array();
		while ($row = mssql_fetch_row( $cur )) {
			$array[] = $row[$numinarray];
		}
		$this->_freeresult( $cur );
		return $array;
	}
	/**
	* @return The first row of the query.
	*/
	function loadRow()
	{
		if (!($cur = $this->query())) {
			return null;
		}
		$ret = null;
		if ($row = mssql_fetch_row( $cur )) {
			$ret = $row;
		}
		$this->_freeresult( $cur );
		return $ret;
	}
	/**
	* Load a list of database rows (numeric column indexing)
	* @param string The field name of a primary key
	* @return array If <var>key</var> is empty as sequential list of returned records.
	* If <var>key</var> is not empty then the returned array is indexed by the value
	* the database key.  Returns <var>null</var> if the query fails.
	*/
	function loadRowList( $key=null )
	{
		if (!($cur = $this->query())) {
			return null;
		}
		$array = array();
		while ($row = mssql_fetch_row( $cur )) {
			if ($key !== null) {
				$array[$row[$key]] = $row;
			} else {
				$array[] = $row;
			}
		}
		$this->_freeresult( $cur );
		return $array;
	}
	/**
	* Quote an identifier name (field, table, etc)
	* @param string 	The identifier
	* @return string 	The quoted identifier
	*/
	function nameQuote( $s )
	{
		$q = str_pad('',2,$this->_nameQuote);
		$parts = explode('.',$s);
		foreach ( $parts as $k=>$v ) {
			$quoted[$k] = $q{0} . trim(trim($v,$q)) . $q{1};
		}
		return implode( '.', $quoted );
	}
	/**
	* Execute the query
	* @return mixed A database resource if successful, FALSE if not.
	*/
	function query()
	{
		if (!is_resource($this->_resource)) {
			return false;
		}

		// set limit and offset
		if ($this->_limit > 0 || $this->_offset > 0) {
			$typ = strtoupper(substr($this->_sql,0,6));
			if ($typ=='SELECT') {
				$expr = '/^\s*(SELECT\s+(?:(?:ALL|DISTINCT)\s+)*)(TOP\s+[0-9]+[%]*)/is';
				$top = '$1 TOP ' . ($this->_offset+$this->_limit);
				$this->_sql = preg_replace($expr,$top,$this->_sql);
			}
		}
		
		if ($this->_debug) {
			$this->_ticker++;
			$this->_log[] = $this->_sql;
		}
		$this->_errorNum = 0;
		$this->_errorMsg = '';
		$this->_cursor = mssql_query( $this->_sql, $this->_resource );

		if (!$this->_cursor) {
			$this->_errorMsg = mssql_get_last_message() . '<br />SQL=' . $this->_sql;

			if ($this->_debug) {
				JError::raiseError('joomla.database:'.$this->_errorNum, 'JDatabaseMSSQL::query: '.$this->_errorMsg );
			}
			return false;
		}
		else {
			if ($this->_offset>0) {
				mssql_data_seek($this->_cursor,$this->_offset);
			}
			return $this->_cursor;
		}
	}
	/**
	* Execute a batch query
	* @return mixed A database resource if successful, FALSE if not.
	*/
	function queryBatch( $abort_on_error=true, $p_transaction_safe = false)
	{
		$this->_errorNum = 0;
		$this->_errorMsg = '';
		if ($p_transaction_safe) {
			// not yet implemented
		}
		$query_split = preg_split ("/[;]+/", $this->_sql);
		$error = 0;
		foreach ($query_split as $command_line) {
			$command_line = trim( $command_line );
			if ($command_line != '') {
				$this->_cursor = mssql_query( $command_line, $this->_resource );
				if (!$this->_cursor) {
					$error = 1;
					$this->_errorNum .= '0 ';
					$this->_errorMsg .= mssql_get_last_message(). ' SQL=' . $this->_sql . '<br />';
					if ($abort_on_error) {
						return $this->_cursor;
					}
				}
			}
		}
		return $error ? false : true;
	}
	/**
	 * Update an Object/Array in the database
	 * @param string	The table name
	 * @param mixed		The Object or an Associative Array
	 * @param string	The Primary Key name
	 * @param bool		Update NULL values
	 * @return mixed	A database resource if successful, FALSE if not.
	 */
	function updateObject( $table, &$object, $keyName, $updateNulls=true )
	{
		$fmtsql = "UPDATE $table SET %s WHERE %s";
		$array = (array)$object;
		$tmp = array();
		$where = '';
		foreach ($array as $k => $v)
		{
			// 1: Do not update the Primary Key (but use it in the WHERE clause)
			// 2: Do not update Arrays, Objects or Reserved Fields
			// 3: Update Standard values
			// 4: Update NULL values allowed
			// 5: Update NULL values not allowed
			if( $k == $keyName ) {
				$where = $this->nameQuote( $keyName ) . '=' . $this->Quote( $v );
				continue;
			} else if( is_array($v) or is_object($v) or $k[0] == '_' ) {
				continue;
			} else if ($v !== null)	{
				$val = $this->isQuoted( $k ) ? $this->Quote( $v ) : $v;
			} else if ($updateNulls) {
					$val = 'NULL';
			} else {
				continue;
			}
			$tmp[] = $this->nameQuote( $k ) . '=' . $val;
		}
		// we have no WHERE clause - the key is missing or incorrect - try to rescue the query
		if (!$where) {
			$keyName = $this->getTableKeys( $table );
			if (isset($object[$keyName])) {
				$where = $this->nameQuote( $keyName ) . '=' . $this->Quote( $object[$keyName] );
			} else {
				return false;
			}
		}
		// run the query
		$this->setQuery( sprintf( $fmtsql, implode( ",", $tmp ) , $where ) );
		return $this->query();
	}





	/**
	 * Free the query result memory and store the row count
	 * @param int The Query Cursor
	 */
	function _freeresult( $cur )
	{
		$this->_numrows[$cur] = mssql_num_rows($cur) - $this->_offset;
		mssql_free_result($cur);
	}
	/**
	 * Add a field or array of field names to the list that are to be returned as dates
	 * @param mixed Field name or array of names
	 * @param string Table containing the field(s)
	 */
	function addDate( $fields, $table='*ALL*' )
	{
		if (is_string( $fields )) {
			$this->_dates[$table][$fields] = $fields;
		} else {
			foreach ( $fields as $v ) {
				$this->_dates[$table][$v] = $v;
			}
		}
	}
	/**
	 * Add a field or array of fields that are of type 'ntext'
	 * @param mixed Field or array of fields
	 * @param string Table containing the field(s)
	 */
	function addNtext( $fields, $table='*ALL*' )
	{
		if (is_string( $fields )) {
			$this->_ntext[$table][$fields] = $fields;
		} else if (is_array($fields)) {
			foreach ( $fields as $v ) {
				$this->_ntext[$table][$v] = $v;
			}
		}
	}
	/**
	* Quote and CAST a date fieldname
	* @param string 	The field name
	* @param string 	The AS alias (default is NULL which uses the field name)
	* @return string 	The quoted field name
	*/
	function dateQuote( $s, $as = null )
	{
		$s = $this->nameQuote($s);
		if (is_null($as)) {
			$as =  ' AS ' . $s;
		} else if ($as) {
			$as = ' AS ' . $as;
		}
		return 'CAST(CONVERT(varchar,' . $s .',120) AS varchar)' . $as;
	}
	/**
	* Get the Primary Key for one or more tables
	* @param mixed		A table name or array list of table names
	* @param bool 		Get all the keys (default is PK only)
	* @return mixed 	The Key for one table, or an array list ([table]=Key)
	*/
	function getTableKey( $tables )
	{
		$return = array();
		foreach ((array)$tables as $tbl) {
			$tbl = strtolower($tbl);
			if ( (!isset($this->_structure['keys'][$tbl])) || ($reload) ) {
				$this->setQuery( 'EXEC sp_pkeys @table_name=\'' . $tbl . '\'' );
				$data = $this->loadAssoc('INDEX_NAME');
				$this->_structure['keys'][$tbl] = $data['COLUMN_NAME'];
			}
			$return[$tbl] = $this->_structure['keys'][$tbl];
		}
		return (is_array($tables)) ? $return : $return[strtolower($tables)];
	}
	/**
	* Get the Indexes for one or more tables
	* @param mixed		A table name or array list of table names
	* @param bool 		Get all the keys (default is PK only)
	* @return array 	Indexes for one table, or array of Indexes keyed by table
	*/
	function getTableIndexes( $tables )
	{
		$return = array();
		foreach ((array)$tables as $tbl) {
			$tbl = strtolower($tbl);
			if ( (!isset($this->_structure['indexes'][$tbl])) || ($reload) ) {
				$this->setQuery( 'EXEC sp_statistics @table_name=\'' . $tbl . '\'' );
				$this->_structure['indexes'][$tbl] = $this->loadAssocList('INDEX_NAME');
			}
			$return[$tbl] = $this->_structure['indexes'][$tbl];
		}
		return (is_array($tables)) ? $return : $return[strtolower($tables)];
	}
	/**
	* Get complete field info for one or more tables
	* @param array 		A list of table names
	* @param boolean 	Whether to reload the fields
	* @return array 	An array of fields by table
	*/
	function getTableFieldsExtended( $tables, $types=true, $reload=false )
	{
		$return = array();
		foreach ((array)$tables as $tbl) {
			$tbl = strtolower($tbl);
			if ( (!isset($this->_structure['fields'][$tbl])) || ($reload) ) {
				$this->setQuery( 'EXEC sp_columns @table_name=\'' . $tbl . '\'' );
				$this->_structure['fields'][$tbl] = $this->loadAssocList('COLUMN_NAME');
			}
			$return[$tbl] = $this->_structure['fields'][$tbl];
			// store the types
			foreach ( $return[$tbl] as $k=>$v ) {
				// quoted, date, ntext
				switch ($v['DATA_TYPE']) {
					case 11:
						$this->addQuoted($k);
						$this->addDate($k,$tbl);
						break;
					case -10:
						$this->addNtext($k,$tbl);
						$this->addQuoted($k);
						break;
					case -9:
						$this->addQuoted($k);
				}
				// primary key
				if (strpos($v['TYPE_NAME'],'identity')) {
					$this->setTableKey($tbl,$k);
				}
			}
		}
		return $return;
	}
	/**
	* Get table names for the current database
	* @param boolean 	Whether to reload the fields
	* @return array 	A list of table names
	*/
	function getTableListExtended( $reload=false )
	{
		if ( (!count($this->_structure['tables'])) || ($reload) ) {
			$this->setQuery( 'EXEC sp_tables @table_type="\'TABLE\'"' );
			foreach ( $this->loadAssocList('TABLE_NAME') as $k=>$v ) {
			 	$this->_structure['tables'][strtolower($k)] = $v;
			}
		}
		return $this->_structure['tables'];
	}
	/**
	 * Check if there are date fields
	 * @param string The table (optional)
	 * @return bool
	 */
	function hasDates( $tbl='*ALL*' )
	{
		return ( (in_array('*ALL',$this->_dates)) || (in_array($tbl,$this->_dates)) );
	}
	/**
	 * Check if there are ntext fields
	 * @param string The table (optional)
	 * @return bool
	 */
	function hasNtext( $tbl='*ALL*' )
	{
		return ( (in_array('*ALL',$this->_ntext)) || (in_array($tbl,$this->_ntext)) );
	}
	/**
	 * Check if field type is a date
	 * @param string The field name
	 * @param string The field's table (optional)
	 * @return bool
	 */
	function isDate( $f, $tbl=null )
	{
		return ( (isset($this->_dates['*ALL*'][$f])) || (isset($this->_dates[$tbl][$f])) );
	}
	/**
	 * Check if field type is ntext
	 * @param string The field name
	 * @param string The field's table (optional)
	 * @return bool
	 */
	function isNtext( $f, $tbl=null )
	{
		return ( (isset($this->_ntext['*ALL*'][$f])) || (isset($this->_ntext[$tbl][$f])) );
	}
	/**
	* Quote and CAST an ntext fieldname
	* @param string 	The field name
	* @param string 	The AS alias (default is NULL which uses the field name)
	* @return string 	The quoted field name
	*/
	function ntextQuote( $s, $as = null )
	{
		$s = $this->nameQuote($s);
		if (is_null($as)) {
			$as =  ' AS ' . $s;
		} else if ($as) {
			$as = ' AS ' . $as;
		}
		return 'CAST(' . $s .' AS text)' . $as;
	}
	/**
	* Build a list of quoted fields for use in a SELECT clause
	* @param string		The table name
	* @param mixed	 	An array of fields, or '*' to load them all
	* @return array 	An array of fields by table
	*/
	function querySelectFields( $tbl, $fields='*' )
	{
		$fields_list = array();
		// 1: A list of fields
		// 2: All the fields from the table
		// 3: Just one field
		if (is_array($fields)) {
			$fields_list = $fields;
		} else if ( strpos($fields,'*') !== false ) {
			foreach ( array_pop($this->getTableFields($tbl)) as $k=>$v ) {
				$fields_list[] = str_replace('*',$k,$fields);
			}
		} else {
			$fields_list[] = $fields;
		}

		$return = array();
		foreach ( $fields_list as $v ) {
			if ($this->isDate($v,$tbl)) {
				$return[] = $this->dateQuote($v);
			} else if ($this->isNtext($v,$tbl)) {
				$return[] = $this->ntextQuote($v);
			} else {
				$return[] = $this->nameQuote($v);
			}
		}
		return implode(', ', $return);
	}
	/**
	* Set the Primary Key for a table
	* @param mixed		The table name or an associative array (table=field)
	* @param string		The field name (ignored if $tbl is an array)
	*/
	function setTableKey( $tbl, $fld='ID' )
	{
		if (is_array($tbl)) {
			$this->_structure['keys'] = array_merge($this->_structure['keys'],$tbl);
		} else {
			$this->_structure['keys'][strtolower($tbl)] = $fld;
		}
	}

}
?>